@@ -41,6 +41,7 @@ sd_section("UI Inputs",
     "numericInput",
     "radioButtons",
     "selectInput",
+    "varSelectInput",
     "sliderInput",
     "submitButton",
     "textInput",
@@ -104,6 +105,7 @@ sd_section("Rendering functions",
   "Functions that you use in your application's server side code, assigning them to outputs that appear in your user interface.",
   c(
     "renderPlot",
+    "renderCachedPlot",
     "renderText",
     "renderPrint",
     "renderDataTable",
@@ -195,7 +197,9 @@ sd_section("Utility functions",
     "exprToFunction",
     "installExprFunction",
     "parseQueryString",
+    "getCurrentOutputInfo",
     "plotPNG",
+    "sizeGrowthRatio",
     "exportTestValues",
     "setSerializer",
     "snapshotExclude",
@@ -206,7 +210,10 @@ sd_section("Utility functions",
     "shinyDeprecated",
     "serverInfo",
     "shiny-options",
-    "onStop"
+    "onStop",
+    "diskCache",
+    "memoryCache",
+    "key_missing"
   )
 )
 sd_section("Plot interaction",
